@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2014 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2015 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -33,6 +33,7 @@
 #include "ResMgr.h"
 #include "log.h"
 #include "url.h"
+#include "HttpAuth.h"
 #include "HttpDir.h"
 #include "misc.h"
 #include "buffer_ssl.h"
@@ -167,6 +168,11 @@ void Http::Init()
    no_cache_this=false;
    no_cache=false;
 
+   auth_sent=0;
+   proxy_auth_sent=0;
+   auth_scheme=HttpAuth::NONE;
+   proxy_auth_scheme=HttpAuth::NONE;
+
    use_propfind_now=true;
    allprop="";
    allprop_len=0;
@@ -204,8 +210,11 @@ void Http::MoveConnectionHere(Http *o)
    conn->ResumeInternal();
    rate_limit=o->rate_limit.borrow();
    last_method=o->last_method; o->last_method=0;
+   last_uri.move_here(o->last_uri);
+   last_url.move_here(o->last_url);
    timeout_timer.Reset(o->timeout_timer);
    state=CONNECTED;
+   tunnel_state=o->tunnel_state;
    o->Disconnect();
    ResumeInternal();
 }
@@ -227,6 +236,8 @@ void Http::DisconnectLL()
 	 SetError(STORE_FAILED,0);
    }
    last_method=0;
+   last_uri.unset();
+   last_url.unset();
    ResetRequestData();
    state=DISCONNECTED;
 }
@@ -287,6 +298,10 @@ void Http::Close()
    }
    array_send=0;
    no_cache_this=false;
+   auth_sent=0;
+   proxy_auth_sent=0;
+   auth_scheme=HttpAuth::NONE;
+   proxy_auth_scheme=HttpAuth::NONE;
    no_ranges=false;
    use_propfind_now=QueryBool("use-propfind",hostname);
    special=HTTP_NONE;
@@ -306,11 +321,25 @@ void Http::Send(const char *format,...)
    conn->send_buf->Put(str);
 }
 
+void Http::Send(const HttpHeader *hdr)
+{
+   Send("%s: %s\r\n",hdr->GetName(),hdr->GetValue());
+}
+
+void Http::AppendHostEncoded(xstring& buf,const char *host)
+{
+   if(is_ipv6_address(host))
+      buf.append('[').append(host).append(']');
+   else
+      buf.append_url_encoded(host,URL_HOST_UNSAFE);
+}
+
 void Http::SendMethod(const char *method,const char *efile)
 {
    xstring& stripped_hostname=xstring::get_tmp(hostname);
    stripped_hostname.truncate_at('%');
-   xstring& ehost=url::encode(xidna_to_ascii(stripped_hostname),URL_HOST_UNSAFE);
+   xstring ehost;
+   AppendHostEncoded(ehost,xidna_to_ascii(stripped_hostname));
    if(portname) {
       ehost.append(':');
       ehost.append(url::encode(portname,URL_PORT_UNSAFE));
@@ -345,6 +374,12 @@ void Http::SendMethod(const char *method,const char *efile)
    if(*efile=='\0')
       efile="/";
 
+   last_uri.set(efile+(proxy?url::path_index(efile):0));
+   if(last_uri.length()==0)
+      last_uri.set("/");
+   if(proxy)
+      last_url.set(efile);
+
    Send("%s %s HTTP/1.1\r\n",method,efile);
    Send("Host: %s\r\n",ehost.get());
    if(user_agent && user_agent[0])
@@ -405,14 +440,31 @@ void Http::SendBasicAuth(const char *tag,const char *user,const char *pass)
    SendBasicAuth(tag,xstring::cat(user,":",pass,NULL));
 }
 
+void Http::SendProxyAuth()
+{
+   proxy_auth_scheme=HttpAuth::NONE;
+   if(!proxy || !proxy_user || !proxy_pass)
+      return;
+   HttpAuth *auth=HttpAuth::Get(HttpAuth::PROXY,GetFileURL(file,NO_USER),proxy_user);
+   if(auth && auth->Update(last_method,last_url)) {
+      proxy_auth_sent++;
+      Send(auth->GetHeader());
+      return;
+   }
+}
+
 void Http::SendAuth()
 {
-   if(proxy && proxy_user && proxy_pass)
-      SendBasicAuth("Proxy-Authorization",proxy_user,proxy_pass);
-   if(user && pass && !(hftp && !QueryBool("use-authorization",proxy)))
+   if(!auth_scheme && hftp && user && pass && QueryBool("use-authorization",proxy)) {
       SendBasicAuth("Authorization",user,pass);
-   else if(!hftp)
-      SendBasicAuth("Authorization",Query("authorization",hostname));
+      return;
+   }
+   auth_scheme=HttpAuth::NONE;
+   HttpAuth *auth=HttpAuth::Get(HttpAuth::WWW,GetFileURL(file,NO_USER),user);
+   if(auth && auth->Update(last_method,last_uri,0)) {
+      auth_sent++;
+      Send(auth->GetHeader());
+   }
 }
 void Http::SendCacheControl()
 {
@@ -478,12 +530,38 @@ void Http::DirFile(xstring& path,const xstring& ecwd,const xstring& efile) const
       path.append('/');
       path.append(efile);
    } else {
+      size_t min_len=path.length()+1;
       if(ecwd[0]!='/')
 	 path.append('/');
       path.append(ecwd);
       if(ecwd.last_char()!='/' && efile.length()>0)
 	 path.append('/');
-      path.append(efile);
+
+      // reduce . and .. at beginning of efile:
+      //  * get the minimum path length (so that we don't remove ~user)
+      //  * skip .; handle .. using basename_ptr to chomp the path.
+      if(path[min_len]=='~') {
+	 while(path[min_len] && path[min_len]!='/')
+	    ++min_len;
+	 if(path[min_len]=='/')
+	    ++min_len;
+      }
+      const char *e=efile;
+      while(e[0]=='.') {
+	 if(e[1]=='/' || e[1]==0)
+	    ++e;
+	 else if(e[1]=='.' && (e[2]=='/' || e[2]==0)) {
+	    if(path.length()<=min_len)
+	       break;
+	    const char *bn=basename_ptr(path+min_len);
+	    path.truncate(bn-path);
+	    e+=2;
+	 } else
+	    break;
+	 if(*e=='/')
+	    ++e;
+      }
+      path.append(e);
    }
 
    // remove "/~" or "/~/"
@@ -564,7 +642,7 @@ void Http::SendRequest(const char *connection,const char *f)
 	 }
 	 pfile.append('@');
       }
-      pfile.append(url::encode(hostname,URL_HOST_UNSAFE));
+      AppendHostEncoded(pfile,hostname);
       if(portname)
       {
 	 pfile.append(':');
@@ -727,6 +805,8 @@ void Http::SendRequest(const char *connection,const char *f)
 	 Send("Destination: %s\r\n",GetFileURL(file1));
       }
    }
+   if(proxy && !https)
+      SendProxyAuth();
    SendAuth();
    if(no_cache || no_cache_this)
       Send("Pragma: no-cache\r\n"); // for HTTP/1.0 compatibility
@@ -830,31 +910,6 @@ void Http::ProceedArrayInfo()
    }
 }
 
-static const char *extract_quoted_header_value(const char *value)
-{
-   static xstring value1;
-   if(*value=='"')
-   {
-      value++;
-      value1.set(value);
-      char *store=value1.get_non_const();
-      while(*value && *value!='"')
-      {
-	 if(*value=='\\' && value[1])
-	    value++;
-	 *store++=*value++;
-      }
-      *store=0;
-   }
-   else
-   {
-      int end=strcspn(value,"()<>@,;:\\\"/[]?={} \t");
-      value1.set(value);
-      value1.truncate(end);
-   }
-   return value1;
-}
-
 void Http::HandleHeaderLine(const char *name,const char *value)
 {
    // use a perfect hash
@@ -981,7 +1036,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       const char *filename=strstr(value,"filename=");
       if(!filename)
 	 return;
-      filename=extract_quoted_header_value(filename+9);
+      filename=HttpHeader::extract_quoted_value(filename+9);
       SetSuggestedFileName(filename);
       return;
    }
@@ -990,11 +1045,53 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       const char *cs=strstr(value,"charset=");
       if(cs)
       {
-	 cs=extract_quoted_header_value(cs+8);
+	 cs=HttpHeader::extract_quoted_value(cs+8);
 	 entity_charset.set(cs);
       }
       return;
    }
+   case_hh("WWW-Authenticate",'W') {
+      const char *user=this->user;
+      const char *pass=this->pass;
+      if(!user || !pass) {
+	 // try auth info from http:authorization setting
+	 const char *auth_c=Query("authorization",hostname);
+         if(auth_c && *auth_c) {
+	    char *auth=alloca_strdup(auth_c);
+	    char *colon=strchr(auth,':');
+	    if(colon) {
+	       *colon=0;
+	       user=auth;
+	       pass=colon+1;
+	    }
+	 }
+      }
+      if(user && pass) {
+	 // FIXME: keep a request queue, get the URI from the queue.
+	 const char *uri=GetFileURL(file,NO_USER);
+
+	 Ref<HttpAuth::Challenge> chal(new HttpAuth::Challenge(value));
+	 bool stale=chal->GetParam("stale").eq_nc("true");
+	 if(status_code==H_Unauthorized) {
+	    if(auth_sent>(stale?1:0))
+	       return;
+	    HttpAuth::scheme_t new_scheme=chal->GetSchemeCode();
+	    if(new_scheme<=auth_scheme)
+	       return;
+	    if(HttpAuth::New(HttpAuth::WWW,uri,chal.borrow(),user,pass))
+	       auth_scheme=new_scheme;
+	 } else if(status_code==H_Proxy_Authentication_Required) {
+	    if(proxy_auth_sent>(stale?1:0))
+	       return;
+	    HttpAuth::scheme_t new_scheme=chal->GetSchemeCode();
+	    if(new_scheme<=proxy_auth_scheme)
+	       return;
+	    if(HttpAuth::New(HttpAuth::PROXY,uri,chal.borrow(),proxy_user,proxy_pass))
+	       proxy_auth_scheme=new_scheme;
+	 }
+      }
+      return;
+   }
    default:
       break;
    }
@@ -1088,23 +1185,30 @@ int Http::Do()
       }
       if(propfind->Eof())
       {
+	 LogNote(9,"got EOF on PROPFIND reply");
 	 const char *b;
 	 int len;
 	 propfind->Get(&b,&len);
-	 FileSet *fs=HttpListInfo::ParseProps(b,len,GetCwd());
+	 Ref<FileSet> fs(HttpListInfo::ParseProps(b,len,GetCwd()));
 	 propfind=0;
 	 if(fs)
 	 {
-	    if(mode==ARRAY_INFO)
-	       fileset_for_info->Merge(fs);
-	    FileInfo *fi=fs->curr();
-	    if(fi && fi->Has(fi->TYPE))
+	    if(mode==CHANGE_DIR)
 	    {
-	       if(mode==CHANGE_DIR)
+	       fs->rewind();
+	       FileInfo *fi=fs->curr();
+	       if(fi && fi->Has(fi->TYPE))
+	       {
+		  LogNote(9,"new-cwd: %s",fi->GetLongName());
 		  new_cwd->is_file=(fi->filetype!=fi->DIRECTORY);
-	       fi->MakeLongName();
-	       LogNote(9,"%s",fi->longname.get());
+		  if(new_cwd->url.last_char()=='/' && new_cwd->is_file)
+		     new_cwd->url.rtrim('/');
+		  else if(new_cwd->url.last_char()!='/' && !new_cwd->is_file)
+		     new_cwd->url.append('/');
+	       }
 	    }
+	    else if(mode==ARRAY_INFO)
+	       fileset_for_info->Merge(fs);
 	 }
 	 m=MOVED;
 	 state=DONE;
@@ -1166,7 +1270,7 @@ int Http::Do()
 	    while(*scan==' ')
 	       scan++;
 	    file_url.set(https?"https://":"http://");
-	    file_url.append_url_encoded(hostname,URL_HOST_UNSAFE);
+	    AppendHostEncoded(file_url,hostname);
 	    if(portname)
 	    {
 	       file_url.append(':');
@@ -1235,7 +1339,7 @@ int Http::Do()
 	 return m;
 
       if(!NextTry())
-      	 return MOVED;
+	 return MOVED;
 
       retry_after=0;
 
@@ -1309,10 +1413,13 @@ int Http::Do()
 	 if(proxy && https)
 	 {
 	    // have to setup a tunnel.
-	    const char *ehost=url::encode(hostname,URL_HOST_UNSAFE);
+	    xstring ehost;
+	    AppendHostEncoded(ehost,hostname);
 	    const char *port_to_use=portname?portname.get():HTTPS_DEFAULT_PORT;
 	    const char *eport=url::encode(port_to_use,URL_PORT_UNSAFE);
-	    Send("CONNECT %s:%s HTTP/1.1\r\n\r\n",ehost,eport);
+	    Send("CONNECT %s:%s HTTP/1.1\r\n",ehost.get(),eport);
+	    SendProxyAuth();
+	    Send("\r\n");
 	    tunnel_state=TUNNEL_WAITING;
 	    state=RECEIVING_HEADER;
 	    return MOVED;
@@ -1429,11 +1536,11 @@ int Http::Do()
 	       {
 		  chunked_trailer=false;
 		  chunked=false;
-		  if(propfind)
+		  if(propfind) {
+		     // we avoid the DONE state since we have yet to handle propfind data
 		     propfind->PutEOF();
-		  if(mode==ARRAY_INFO)
 		     state=CONNECTED;
-		  else
+		  } else
 		     state=DONE;
 		  return MOVED;
 	       }
@@ -1597,6 +1704,16 @@ int Http::Do()
 	 return MOVED;
       }
 
+      if((status_code==H_Unauthorized && auth_scheme)
+      || (status_code==H_Proxy_Authentication_Required && proxy_auth_scheme)) {
+	 // retry with authentication
+	 state=RECEIVING_BODY;
+	 LogErrorText();
+	 Disconnect();
+	 DontSleep();
+	 return MOVED;
+      }
+
       if(!H_2XX(status_code))
       {
 	 xstring err;
@@ -1604,7 +1721,8 @@ int Http::Do()
 
 	 if(H_REDIRECTED(status_code))
 	 {
-	    if(location && !url::is_url(location)
+	    bool is_url=(location && url::is_url(location));
+	    if(location && !is_url
 	    && mode==QUOTE_CMD && !strncasecmp(file,"POST ",5)
 	    && tunnel_state!=TUNNEL_WAITING)
 	    {
@@ -1637,6 +1755,14 @@ int Http::Do()
 		  strcpy(slash+1,location);
 	       }
 	       location.set(new_location);
+	    } else if(is_url && !hftp) {
+	       ParsedURL url(location);
+	       if(url.proto.eq(GetProto()) && !xstrcasecmp(url.host,hostname)
+	       && user && !url.user) {
+		  // use the same user name after redirect to the same site.
+		  url.user.set(user);
+		  location.set_allocated(url.Combine());
+	       }
 	    }
 	    err.setf("%s (%s -> %s)",status+status_consumed,file.get(),
 				    location?location.get():"nowhere");
@@ -1684,8 +1810,10 @@ int Http::Do()
 	 return MOVED;
       }
       if(!xstrcmp(last_method,"PROPFIND")
-      && (mode==ARRAY_INFO || mode==CHANGE_DIR))
+      && (mode==ARRAY_INFO || mode==CHANGE_DIR)) {
+	 LogNote(9,"accepting XML for PROPFIND...");
 	 propfind=new IOBufferFileAccess(this);
+      }
 
       if(mode==CHANGE_DIR && !propfind)
       {
@@ -1924,6 +2052,7 @@ get_again:
 	 chunked_trailer=true;
 	 state=RECEIVING_HEADER;
 	 body_size=bytes_received;
+	 Timeout(0);
 	 return DO_AGAIN;
       }
       if(chunk_pos==chunk_size)
@@ -1965,6 +2094,8 @@ get_again:
 	 if(chunked)
 	    chunk_pos+=size1;
 	 bytes_received+=size1;
+	 if(inflate->Error())
+	    SetError(FATAL,inflate->ErrorText());
       }
       inflate->Get(&buf1,&size1);
       src_buf=inflate.get_non_const();
@@ -2072,7 +2203,7 @@ int Http::SendEOT()
 	    keep_alive=false;
 	 }
 	 sent_eot=true;
-      	 return(OK);
+	 return(OK);
       }
       return(DO_AGAIN);
    }
@@ -2349,7 +2480,7 @@ void Http::SetCookie(const char *value_const)
       if(!strncasecmp(entry,"path=",5))
       {
 	 path=alloca_strdup(entry+5);
-      	 continue;
+	 continue;
       }
 
       if(!strncasecmp(entry,"domain=",7))
@@ -2600,6 +2731,8 @@ bool Http::CompressedContentEncoding() const
 }
 bool Http::CompressedContentType() const
 {
+   if(file.ends_with(".gz") || file.ends_with(".Z") || file.ends_with(".tgz"))
+      return true;
    static const char app[]="application/";
    return entity_content_type && entity_content_type.begins_with(app)
       && IsCompressed(entity_content_type+sizeof(app)-1);
