@@ -17,8 +17,6 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/* $Id: PatternSet.cc,v 1.6 2008/11/27 05:56:23 lav Exp $ */
-
 #include <config.h>
 #include <fnmatch.h>
 #include "PatternSet.h"
@@ -26,11 +24,28 @@
 PatternSet::PatternSet()
 {
    chain=0;
+   first=0;
 }
 void PatternSet::Add(Type t,Pattern *p)
 {
    chain=new PatternLink(t,p,chain);
+   if(!first)
+      first=chain;
 }
+
+void PatternSet::AddFirst(Type t,Pattern *p)
+{
+   PatternLink *new_link=new PatternLink(t,p,NULL);
+   if(!first)
+      first=chain=new_link;
+   else
+      first->next=new_link;
+}
+PatternSet::Type PatternSet::GetFirstType() const
+{
+   return first->type;
+}
+
 PatternSet::~PatternSet()
 {
    while(chain)
@@ -89,8 +104,10 @@ PatternSet::Regex::Regex(const char *p) : Pattern(p)
    if(errcode)
    {
       size_t need=regerror(errcode,0,0,0);
-      error.get_space(need-1);
-      error.set_length(regerror(errcode,0,error.get_non_const(),need)-1);
+      xstring& e=xstring::get_tmp();
+      e.get_space(need-1);
+      e.set_length(regerror(errcode,0,e.get_non_const(),need)-1);
+      error.setf(_("regular expression `%s': %s"),p,e.get());
    }
 }
 PatternSet::Regex::~Regex()
